"""
Utilities for testing.
"""

# data examples for use with pytest parameterized tests
# (creating a data value here because it's simpler to use
# than a fixture inside parameterized tests).
example_data_for_image_tests = [
    # test example data without labels
    (
        "tests/data/random_tiff_z_stacks/Z99",
        None,
        "output.zarr",
        {
            "111": "Channel A",
            "222": "Channel B",
            "333": "Channel C",
            "444": "Channel D",
            "555": "Channel E",
        },
        (1.0, 0.1, 0.1),
        None,
        None,
    ),
    # test example data with labels
    (
        "tests/data/random_tiff_z_stacks/Z99",
        "tests/data/random_tiff_z_stacks/labels",
        "output.zarr",
        {
            "111": "Channel A",
            "222": "Channel B",
            "333": "Channel C",
            "444": "Channel D",
            "555": "Channel E",
        },
        (1.0, 0.1, 0.1),
        None,
        ["compartment (labels)"],
    ),
    # test example data with labels and ignore a channel
    (
        "tests/data/random_tiff_z_stacks/Z99",
        "tests/data/random_tiff_z_stacks/labels",
        "output.zarr",
        {
            "111": "Channel A",
            "222": "Channel B",
            "333": "Channel C",
            "444": "Channel D",
            "555": "Channel E",
        },
        (1.0, 0.1, 0.1),
        ["555"],
        ["compartment (labels)"],
    ),
]



"""
Tests for the image module.
"""

import pathlib
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import pytest
import tifffile as tiff
import zarr

from nviz.image import image_set_to_arrays, tiff_to_ometiff, tiff_to_zarr
from tests.utils import example_data_for_image_tests


@pytest.mark.parametrize(
    (
        "image_dir, label_dir, output_path, channel_map, "
        "scaling_values, ignore, expected_labels"
    ),
    example_data_for_image_tests,
)
def test_image_set_to_arrays(
    image_dir: str,
    label_dir: Optional[str],
    output_path: str,
    channel_map: Dict[str, str],
    scaling_values: Tuple[int, int, int],
    ignore: Optional[List[str]],
    expected_labels: List[str],
):
    # Call the function
    result = image_set_to_arrays(
        image_dir=image_dir, label_dir=label_dir, channel_map=channel_map, ignore=ignore
    )

    # check that we have all keys
    if ignore is None:
        all(channel not in result["images"] for channel in channel_map.values())

    # check that we ignored what we should have
    elif ignore is not None:
        assert all(ignored not in result["images"] for ignored in ignore)


@pytest.mark.parametrize(
    (
        "image_dir, label_dir, output_path, channel_map, "
        "scaling_values, ignore, expected_labels"
    ),
    example_data_for_image_tests,
)
def test_tiff_to_zarr(
    image_dir: str,
    label_dir: Optional[str],
    output_path: str,
    channel_map: Dict[str, str],
    scaling_values: Tuple[int, int, int],
    ignore: Optional[List[str]],
    expected_labels: List[str],
    tmp_path: pathlib.Path,
):
    """
    Tests the tiff_to_zarr function.
    """

    output_path = tiff_to_zarr(
        image_dir=image_dir,
        label_dir=label_dir,
        output_path=f"{tmp_path}/{output_path}",
        channel_map=channel_map,
        scaling_values=scaling_values,
        ignore=ignore,
    )

    # Check if the output path exists
    assert Path(output_path).exists()

    # Check if the Zarr structure is correct
    zarr_root = zarr.open(output_path, mode="r")
    assert "images" in list(zarr_root.keys())

    # check if we have labels if we supplied them
    if label_dir is not None:
        assert "labels" in zarr_root

    for channel in channel_map.values():
        if ignore is not None and channel not in [
            channel_map[ignored] for ignored in ignore
        ]:
            assert channel in list(zarr_root["images"])

    # check if we have labels if we supplied them
    if label_dir is not None:
        assert all(
            expected_label in list(zarr_root["labels"].keys())
            for expected_label in expected_labels
        )


@pytest.mark.parametrize(
    (
        "image_dir, label_dir, output_path, channel_map, "
        "scaling_values, ignore, expected_labels"
    ),
    example_data_for_image_tests,
)
def test_tiff_to_ometiff(
    image_dir: str,
    label_dir: Optional[str],
    output_path: str,
    channel_map: Dict[str, str],
    scaling_values: Tuple[int, int, int],
    ignore: Optional[List[str]],
    expected_labels: List[str],
    tmp_path: pathlib.Path,
):
    """
    Tests the tiff_to_ometiff function.
    """

    output_path = tiff_to_ometiff(
        image_dir=image_dir,
        label_dir=label_dir,
        output_path=f"{tmp_path}/{output_path}",
        channel_map=channel_map,
        scaling_values=scaling_values,
        ignore=ignore,
    )

    # Check if the output path exists
    assert Path(output_path).exists()

    # Read the OME-TIFF file and check its contents
    with tiff.TiffFile(output_path) as tif:
        assert len(tif.pages) > 0
        metadata = tif.ome_metadata
        assert metadata is not None

        # Parse the OME-XML metadata
        root = ET.fromstring(metadata)
        channels = root.find(
            ".//{http://www.openmicroscopy.org/Schemas/OME/2016-06}Pixels"
        ).findall("{http://www.openmicroscopy.org/Schemas/OME/2016-06}Channel")

        # Check the metadata for channels
        for channel in channel_map.values():
            if ignore is not None and channel not in [
                channel_map[ignored] for ignored in ignore
            ]:
                assert any(channel == ch.get("Name") for ch in channels)

        # Check the metadata for physical sizes
        pixels = root.find(
            ".//{http://www.openmicroscopy.org/Schemas/OME/2016-06}Pixels"
        )
        assert pixels.get("PhysicalSizeX") == str(scaling_values[2])
        assert pixels.get("PhysicalSizeY") == str(scaling_values[1])
        assert pixels.get("PhysicalSizeZ") == str(scaling_values[0])



"""
Tests for nviz/view.py
"""

import pathlib
from typing import Dict, List, Optional, Tuple, Union
from xml.etree.ElementTree import fromstring

import pytest

from nviz.image_meta import (
    extract_z_slice_number_from_filename,
    gather_scaling_info_from_scaninfoxml,
    generate_ome_xml,
)


@pytest.mark.parametrize(
    "filename, expected",
    [
        ("C10-1_405_ZS034_FOV-1.tif", 34),
        ("C10-1_405_ZS018_FOV-1.tif", 18),
        ("C10-1_405_ZS039_FOV-1.tif", 39),
        ("C10-1_405_ZS043_FOV-1.tif", 43),
        ("C10-1_405_ZS033_FOV-1.tif", 33),
        ("C10-1_405_ZS027_FOV-1.tif", 27),
        ("C10-1_405_ZS006_FOV-1.tif", 6),
        ("C10-1_405_FOV-1.tif", 0),  # No ZS pattern
        ("C10-1_405_ZS_FOV-1.tif", 0),  # Incomplete ZS pattern
    ],
)
def test_extract_z_slice_number_from_filename(filename: str, expected: int):
    """
    Tests extract_z_slice_number_from_filename
    """
    assert extract_z_slice_number_from_filename(filename) == expected


@pytest.mark.parametrize(
    "xml_content, expected",
    [
        (
            """<?xml version="1.0" encoding="utf-8"?>
            <ScanInfo>
                <Group Name="Calibration">
                    <Settings>
                        <Setting Parameter="MicronsPerPixelX">0.1006</Setting>
                        <Setting Parameter="MicronsPerPixelY">0.1006</Setting>
                    </Settings>
                </Group>
                <Group Name="Experiment">
                    <Settings>
                        <Setting Parameter="ZStackSpacingMicrons">1.000</Setting>
                    </Settings>
                </Group>
            </ScanInfo>""",
            (1.000, 0.1006, 0.1006),
        ),
        (
            """<?xml version="1.0" encoding="utf-8"?>
            <ScanInfo>
                <Group Name="Calibration">
                    <Settings>
                        <Setting Parameter="MicronsPerPixelX">0.200</Setting>
                        <Setting Parameter="MicronsPerPixelY">0.200</Setting>
                    </Settings>
                </Group>
                <Group Name="Experiment">
                    <Settings>
                        <Setting Parameter="ZStackSpacingMicrons">2.000</Setting>
                    </Settings>
                </Group>
            </ScanInfo>""",
            (2.000, 0.200, 0.200),
        ),
        (
            """<?xml version="1.0" encoding="utf-8"?>
            <ScanInfo>
                <Group Name="Calibration">
                    <Settings>
                        <Setting Parameter="MicronsPerPixelX">0.300</Setting>
                    </Settings>
                </Group>
                <Group Name="Experiment">
                    <Settings>
                        <Setting Parameter="ZStackSpacingMicrons">3.000</Setting>
                    </Settings>
                </Group>
            </ScanInfo>""",
            (3.000, None, 0.300),
        ),
    ],
)
def test_gather_scaling_info_from_scaninfoxml(
    xml_content: str,
    expected: Tuple[Optional[float], Optional[float], Optional[float]],
    tmp_path: pathlib.Path,
):
    """
    Tests gather_scaling_info_from_scaninfoxml
    """

    # write a temp file
    xml_file = tmp_path / "temp_scaninfo.xml"
    xml_file.write_text(xml_content)

    # check the results
    assert gather_scaling_info_from_scaninfoxml(xml_file) == expected


@pytest.mark.parametrize(
    "metadata, expected_channels",
    [
        (
            {
                "SizeC": 3,
                "SizeZ": 10,
                "SizeY": 512,
                "SizeX": 512,
                "PhysicalSizeX": 0.1,
                "PhysicalSizeY": 0.1,
                "PhysicalSizeZ": 1.0,
                "PhysicalSizeXUnit": "um",
                "PhysicalSizeYUnit": "um",
                "PhysicalSizeZUnit": "um",
                "Channel": [
                    {"Name": "Channel A"},
                    {"Name": "Channel B"},
                    {"Name": "Channel C"},
                ],
            },
            ["Channel A", "Channel B", "Channel C"],
        ),
        (
            {
                "SizeC": 2,
                "SizeZ": 5,
                "SizeY": 256,
                "SizeX": 256,
                "PhysicalSizeX": 0.2,
                "PhysicalSizeY": 0.2,
                "PhysicalSizeZ": 2.0,
                "PhysicalSizeXUnit": "um",
                "PhysicalSizeYUnit": "um",
                "PhysicalSizeZUnit": "um",
                "Channel": [
                    {"Name": "Channel X"},
                    {"Name": "Channel Y"},
                ],
            },
            ["Channel X", "Channel Y"],
        ),
    ],
)
def test_generate_ome_xml(
    metadata: Dict[str, Union[int, str]], expected_channels: List[str]
):
    ome_xml = generate_ome_xml(metadata)

    # Parse the generated OME-XML
    root = fromstring(ome_xml)

    # Verify the Pixels attributes
    pixels = root.find(".//{http://www.openmicroscopy.org/Schemas/OME/2016-06}Pixels")
    assert pixels is not None
    assert pixels.get("SizeC") == str(metadata["SizeC"])
    assert pixels.get("SizeZ") == str(metadata["SizeZ"])
    assert pixels.get("SizeY") == str(metadata["SizeY"])
    assert pixels.get("SizeX") == str(metadata["SizeX"])
    assert pixels.get("PhysicalSizeX") == str(metadata["PhysicalSizeX"])
    assert pixels.get("PhysicalSizeY") == str(metadata["PhysicalSizeY"])
    assert pixels.get("PhysicalSizeZ") == str(metadata["PhysicalSizeZ"])
    assert pixels.get("PhysicalSizeXUnit") == metadata["PhysicalSizeXUnit"]
    assert pixels.get("PhysicalSizeYUnit") == metadata["PhysicalSizeYUnit"]
    assert pixels.get("PhysicalSizeZUnit") == metadata["PhysicalSizeZUnit"]

    # Verify the Channel names
    channels = pixels.findall(
        "{http://www.openmicroscopy.org/Schemas/OME/2016-06}Channel"
    )
    assert len(channels) == len(expected_channels)
    for i, channel in enumerate(channels):
        assert channel.get("Name") == expected_channels[i]



"""
Tests for the view module
"""

import pathlib
from typing import Dict, List, Optional, Tuple

import pytest

from nviz.image import tiff_to_ometiff, tiff_to_zarr
from nviz.view import view_ometiff_with_napari, view_zarr_with_napari
from tests.utils import example_data_for_image_tests


@pytest.mark.parametrize(
    (
        "image_dir, label_dir, output_path, channel_map, "
        "scaling_values, ignore, expected_labels"
    ),
    example_data_for_image_tests,
)
def test_view_zarr_with_napari(
    image_dir: str,
    label_dir: Optional[str],
    output_path: str,
    channel_map: Dict[str, str],
    scaling_values: Tuple[int, int, int],
    ignore: Optional[List[str]],
    expected_labels: List[str],
    tmp_path: pathlib.Path,
):
    """
    Tests the view_zarr_with_napari function.
    """

    zarr_dir = tiff_to_zarr(
        image_dir=image_dir,
        label_dir=label_dir,
        output_path=f"{tmp_path}/{output_path}",
        channel_map=channel_map,
        scaling_values=scaling_values,
        ignore=ignore,
    )

    # Call the function
    view = view_zarr_with_napari(
        zarr_dir=zarr_dir, scaling_values=scaling_values, headless=True
    )

    # close the view
    view.close()


@pytest.mark.parametrize(
    (
        "image_dir, label_dir, output_path, channel_map, "
        "scaling_values, ignore, expected_labels"
    ),
    example_data_for_image_tests,
)
def test_view_ometiff_with_napari(
    image_dir: str,
    label_dir: Optional[str],
    output_path: str,
    channel_map: Dict[str, str],
    scaling_values: Tuple[int, int, int],
    ignore: Optional[List[str]],
    expected_labels: List[str],
    tmp_path: pathlib.Path,
):
    """
    Tests the view_ometiff_with_napari function.
    """

    ometiff_path = tiff_to_ometiff(
        image_dir=image_dir,
        label_dir=label_dir,
        output_path=f"{tmp_path}/{output_path}",
        channel_map=channel_map,
        scaling_values=scaling_values,
        ignore=ignore,
    )

    # Call the function
    view = view_ometiff_with_napari(
        ometiff_path=ometiff_path,
        scaling_values=scaling_values,
        headless=True,
    )

    # close the view
    view.close()



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.16.6
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # nViz Quickstart Guide
#
# This notebook helps demonstrate how to use `nViz` quickly.
# We use demonstrational data found within this repository
# to help show how the software works.

# +
import pathlib
import shutil

from napari.utils import nbscreenshot

import nviz

# +
# setup paths and prepare filesystem
test_image_dir = "../../../tests/data/random_tiff_z_stacks/Z99/"
test_label_dir = "../../../tests/data/random_tiff_z_stacks/labels/"
output_ome_tiff_path = "./example.ome.tiff"
output_zarr_path = "./example.zarr"
# scaling values in micrometers for z, y, and x dimensions.
# Note: This is typically a priori knowledge that is usually
# found in raw microscopy data (for example raw TIFF files
# or other microscopy metadata files [like XML]).
scaling_values = (1.0, 0.1, 0.1)

# If we already have a demo tiff file, remove it.
# (We are demoing the full nViz process from scratch.)
if (tiff_path := pathlib.Path(output_ome_tiff_path)).is_file():
    tiff_path.unlink()

# If we already have a demo zarr dir, remove it
# (We are demoing the full nViz process from scratch.)
if (zarr_path := pathlib.Path(output_zarr_path)).is_dir():
    shutil.rmtree(output_zarr_path)
# -

# show our images
list(pathlib.Path(test_image_dir).rglob("*.tif"))

# show our label files (in this case, segmentation masks)
# Other labels might include outlines, bounding boxes,
# or similar data stored as images which provide
# information about objects of interest.
list(pathlib.Path(test_label_dir).rglob("*.tif"))

# +
# convert original raw TIFF into OME-TIFF format using nViz

nviz_ome_tiff = nviz.tiff_to_ometiff(
    # specify an image directory (original images)
    image_dir=test_image_dir,
    # specify a label (masks or outlines) directory
    label_dir=test_label_dir,
    # specify an output path
    output_path=output_ome_tiff_path,
    # Map filename patterns to channel names.
    # Note: the keys of this dictionary correspond to information
    # found in the file name. The value pairs are customizable, and
    # they will display as specified here inside a Napari window.
    channel_map={
        "111": "Channel A",
        "222": "Channel B",
        "333": "Channel C",
        "444": "Channel D",
        "555": "Channel E",
    },
    # set the scaling for z, y, and x
    scaling_values=scaling_values,
)

# show the filepath
nviz_ome_tiff
# -

# view the ome-tiff in napari
# (will open a new window and also returns the viewer object)
viewer = nviz.view_ometiff_with_napari(
    ometiff_path=nviz_ome_tiff, scaling_values=scaling_values
)

# show the napari window within this notebook to help demonstrate output
viewer.window.resize(1000, 700)
viewer.camera.angles = (50, 50, 30)
viewer.camera.zoom = 30
nbscreenshot(viewer=viewer, canvas_only=False)

# +
# convert original data into a Zarr which
# includes OME-Zarr images format using nViz

nviz_zarr = nviz.tiff_to_zarr(
    # specify an image directory (original images)
    image_dir=test_image_dir,
    # specify a label (masks or outlines) directory
    label_dir=test_label_dir,
    # specify an output path
    output_path=output_zarr_path,
    # Map filename patterns to channel names.
    # Note: the keys of this dictionary correspond to information
    # found in the file name. The value pairs are customizable, and
    # they will display as specified here inside a Napari window.
    channel_map={
        "111": "Channel A",
        "222": "Channel B",
        "333": "Channel C",
        "444": "Channel D",
        "555": "Channel E",
    },
    # set the scaling for z, y, and x
    scaling_values=scaling_values,
)

# show the filepath
nviz_zarr
# -

# view the zarr in napari
# (will open a new window and also returns the viewer object)
viewer = nviz.view_zarr_with_napari(
    zarr_dir=output_zarr_path, scaling_values=scaling_values
)

# show the napari window within this notebook to help demonstrate output
viewer.window.resize(1000, 700)
viewer.camera.angles = (50, 50, -30)
viewer.camera.zoom = 30
nbscreenshot(viewer=viewer, canvas_only=False)



"""
Utilities for viewing n-dimensional data
"""

import logging
from typing import Optional

import napari
import tifffile as tiff
import xmltodict
import zarr

logger = logging.getLogger(__name__)


def view_zarr_with_napari(
    zarr_dir: str, scaling_values: tuple, headless: bool = False
) -> Optional[napari.Viewer]:
    """
    View a Zarr file created with nviz through napari.

    Args:
        zarr_dir (str):
            The path to the Zarr file.
        scaling_values (tuple):
            The scaling values for the image.
        headless (bool):
            Whether to run in headless mode
            (where we don't run napari and hand
            back the viewer object instead).

    Returns:
        Optional[napari.Viewer]]:
            The napari viewer object if not headless,
            otherwise None.
    """
    # Check Zarr file structure
    frame_zarr = zarr.open(zarr_dir, mode="r")

    # Visualize with napari, start in 3d mode
    viewer = napari.Viewer(ndisplay=3)

    # Iterate through each channel in the Zarr file
    for channel_name in sorted(frame_zarr["images"].keys(), reverse=True):
        viewer.add_image(
            frame_zarr["images"][channel_name]["0"][:],
            name=channel_name,
            scale=scaling_values,
        )

    # Iterate through each compartment in the Zarr file and add labels to Napari
    if "labels" in frame_zarr:
        for label_name in sorted(frame_zarr["labels"].keys(), reverse=True):
            viewer.add_labels(
                frame_zarr["labels"][label_name]["0"][:],
                name=f"{label_name}",
                scale=scaling_values,
            )

    if not headless:
        # Start the Napari event loop
        napari.run()
    else:
        logger.warning(
            "Running view in headless mode and returning a napari viewer object."
        )

    # otherwise return the viewer
    return viewer


def view_ometiff_with_napari(
    ometiff_path: str, scaling_values: tuple, headless: bool = False
) -> Optional[napari.Viewer]:
    """
    View a OME-TIFF file created with nviz through napari.

    Args:
        ometiff_path (str):
            The path to the OME-TIFF file.
        scaling_values (tuple):
            The scaling values for the image.
        headless (bool):
            Whether to run in headless mode
            (where we don't run napari and hand
            back the viewer object instead).

    Returns:
        Optional[napari.Viewer]]:
            The napari viewer object if not headless,
            otherwise None.
    """

    # Visualize with napari, start in 3d mode
    viewer = napari.Viewer(ndisplay=3)

    # Read and add layers from the combined OME-TIFF file
    with tiff.TiffFile(ometiff_path) as tif:
        combined_data = tif.asarray()
        metadata = xmltodict.parse(tif.ome_metadata)
        channel_names = [
            channel["@Name"]
            for channel in metadata["OME"]["Image"]["Pixels"]["Channel"]
        ]

        # First, add image layers
        for i, (channel_data, channel_name) in enumerate(
            zip(combined_data, channel_names)
        ):
            if "(labels)" not in channel_name:
                viewer.add_image(
                    channel_data,
                    name=channel_name,
                    scale=scaling_values,
                )

        # Then, add label layers
        for i, (channel_data, channel_name) in enumerate(
            zip(combined_data, channel_names)
        ):
            if "(labels)" in channel_name:
                viewer.add_labels(
                    channel_data,
                    name=channel_name,
                    scale=scaling_values,
                )

    if not headless:
        # Start the Napari event loop
        napari.run()
    else:
        logger.warning(
            "Running view in headless mode and returning a napari viewer object."
        )

    # otherwise return the viewer
    return viewer



"""
Generates example z-stack TIFF files with randomized data for testing.
"""

import pathlib
from typing import Tuple

import numpy as np
import tifffile as tiff

# human readable channel names
channels = [
    "Channel A",
    "Channel B",
    "Channel C",
    "Channel D",
    "Channel E",
]

# map human readable channel names to machine ones
# which simulate what is produced by the microscope
# or teams who create images.
channels_map = {
    "Channel A": "111",
    "Channel B": "222",
    "Channel C": "333",
    "Channel D": "444",
    "Channel E": "555",
}

# gather a relative path
relpath = pathlib.Path(__file__).parent

# Set common static scaling values
scaling_values = (1.0, 0.1, 0.1)  # (z, y, x) in microns

# Generate random data for z-slices
num_z_slices = 10
image_shape = (100, 100)  # Example shape, adjust as needed

# create random data for each channel
channels = {
    channel: [
        np.random.randint(0, 65535, size=image_shape, dtype=np.uint16)
        for _ in range(num_z_slices)
    ]
    for channel in channels
}

# Debug: show channel keys and file counts
print(channels.keys())
for channel, files in channels.items():
    print(f"Channel: {channel}, Slices: {len(files)}")

# Define the output directory
output_dir = relpath / "Z99"
pathlib.Path(output_dir).mkdir(exist_ok=True)

# Write each z-slice as a separate TIFF file
for channel, z_slices in channels.items():
    for z_index, z_slice in enumerate(z_slices):
        filename = f"Z99_{channels_map[channel]}_ZS{z_index:03d}.tif"
        filepath = output_dir / filename
        tiff.imwrite(filepath, z_slice)

print(f"TIFF files written to {output_dir}")


# Create a single multidimensional TIFF file with sphere shapes
def create_sphere_image(
    shape: Tuple[int, int, int], radius: int, center: Tuple[int, int, int]
) -> np.ndarray:
    """Create a 3D image with a sphere.

    Args:
        shape (Tuple[int, int, int]): The shape of the 3D image (z, y, x).
        radius (int): The radius of the sphere.
        center (Tuple[int, int, int]): The center of the sphere (z, y, x).

    Returns:
        np.ndarray: A 3D image with a sphere.
    """
    z, y, x = np.indices(shape)
    distance = np.sqrt(
        (z - center[0]) ** 2 + (y - center[1]) ** 2 + (x - center[2]) ** 2
    )
    sphere = (distance <= radius).astype(np.uint16) * 65535
    return sphere


sphere_radius = 20
sphere_center = (num_z_slices // 2, image_shape[0] // 2, image_shape[1] // 2)
sphere_image = create_sphere_image(
    (num_z_slices, *image_shape), sphere_radius, sphere_center
)

# create a path for labels
label_path = output_dir.parent / "labels"
pathlib.Path(label_path).mkdir(exist_ok=True)

# Write the sphere image as a single multidimensional TIFF file
sphere_tiff_path = label_path / "compartment.tif"
tiff.imwrite(sphere_tiff_path, sphere_image, photometric="minisblack")

print(f"Sphere TIFF file written to {sphere_tiff_path}")

scaninfo_file = relpath / "ScanInfo.xml"

with open(scaninfo_file, "w") as file:
    file.write(
        """
<?xml version="1.0" encoding="utf-8"?>
<ScanInfo>
  <Version>0.0.0.0</Version>
  <Group Name="Calibration">
    <Settings>
      <Setting Parameter="MicronsPerPixelX">0.1006</Setting>
      <Setting Parameter="MicronsPerPixelY">0.1006</Setting>
    </Settings>
  </Group>
  <Group Name="Experiment">
    <Settings>
      <Setting Parameter="ZStackSpacingMicrons">1.000</Setting>
    </Settings>
  </Group>
</ScanInfo>
"""
    )



"""
Viewing utilities for GFF 3D organoid project
"""

import re
import xml.etree.ElementTree as ET
from typing import Dict, Optional, Tuple
from xml.etree.ElementTree import Element, SubElement, tostring


def extract_z_slice_number_from_filename(filename: str) -> int:
    """
    Extracts a z-slice number from a filename.
    Follows conventions from images which are produced
    to our knowledge by ZEISS LSM 880 with Airyscan
    microscope output.

    Assumes the z-slice number
    is a zero-padded number in the filename
    with the pattern '_ZS###_'.

    Args:
        filename (str):
            The name of the file from which to extract the z-slice number.

    Returns:
        int:
            The extracted z-slice number. Returns 0 if the pattern is not found.
    """
    match = re.search(r"_ZS(\d+)_", filename)
    return int(match.group(1)) if match else 0


def gather_scaling_info_from_scaninfoxml(
    xml_file: str,
) -> Tuple[Optional[float], Optional[float], Optional[float]]:
    """
    Reads the scan information from an XML file and
    returns the values of ZStackSpacingMicrons,
    MicronsPerPixelY, and MicronsPerPixelX.

    This function specifically caters to a file named
    ScanInfo.xml which is included to our knowledge within
    ZEISS LSM 880 with Airyscan microscope output (and perhaps
    others).

    Args:
        xml_file (str):
            Path to the XML file.

    Returns:
        Tuple[Optional[float], Optional[float], Optional[float]]:
            A tuple containing the values of
            ZStackSpacingMicrons, MicronsPerPixelY,
            and MicronsPerPixelX. If a value is not found,
            it will be None.
    """
    tree = ET.parse(xml_file)
    root = tree.getroot()

    microns_per_pixel_y: Optional[float] = None
    microns_per_pixel_x: Optional[float] = None
    z_stack_spacing_microns: Optional[float] = None

    for setting in root.findall(".//Setting"):
        param = setting.get("Parameter")
        if param == "MicronsPerPixelY":
            microns_per_pixel_y = float(setting.text)
        elif param == "MicronsPerPixelX":
            microns_per_pixel_x = float(setting.text)
        elif param == "ZStackSpacingMicrons":
            z_stack_spacing_microns = float(setting.text)

    return (z_stack_spacing_microns, microns_per_pixel_y, microns_per_pixel_x)


def generate_ome_xml(metadata: Dict) -> str:
    """
    Generate OME-XML metadata for use within an OME-TIFF file.

    Args:
        metadata (Dict): Dictionary containing metadata.

    Returns:
        str: OME-XML string.
    """
    ome = Element("OME", xmlns="http://www.openmicroscopy.org/Schemas/OME/2016-06")
    image = SubElement(ome, "Image", ID="Image:0")
    pixels = SubElement(
        image,
        "Pixels",
        {
            "ID": "Pixels:0",
            "Type": "uint16",
            "DimensionOrder": "CZYX",
            "SizeC": str(metadata["SizeC"]),
            "SizeZ": str(metadata["SizeZ"]),
            "SizeY": str(metadata["SizeY"]),
            "SizeX": str(metadata["SizeX"]),
            "PhysicalSizeX": str(metadata["PhysicalSizeX"]),
            "PhysicalSizeY": str(metadata["PhysicalSizeY"]),
            "PhysicalSizeZ": str(metadata["PhysicalSizeZ"]),
            "PhysicalSizeXUnit": metadata["PhysicalSizeXUnit"],
            "PhysicalSizeYUnit": metadata["PhysicalSizeYUnit"],
            "PhysicalSizeZUnit": metadata["PhysicalSizeZUnit"],
        },
    )
    for i, channel in enumerate(metadata["Channel"]):
        SubElement(
            pixels,
            "Channel",
            {
                "ID": f"Channel:0:{i}",
                "Name": channel["Name"],
            },
        )
    return tostring(ome, encoding="unicode")



"""
Experiment with GFF image stacks to OME-ZARR with display in Napari.
"""

import os
import pathlib
from itertools import groupby
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import tifffile as tiff
import zarr
from ome_zarr.io import parse_url as zarr_parse_url
from ome_zarr.writer import write_image as zarr_write_image

from .image_meta import extract_z_slice_number_from_filename, generate_ome_xml


def image_set_to_arrays(
    image_dir: str,
    label_dir: Optional[str],
    channel_map: Dict[str, str],
    ignore: Optional[List[str]] = ["Merge"],
) -> Dict[str, Dict[str, np.ndarray]]:
    """
    Read a set of images as an array of images.
    We follow a convention of splitting the following
    into separate nested dictionaries for use by
    other functions within this project.

    - "images": original images
    - "labels": images which represent objects of interest
        within the original "images"

    Args:
        image_dir (str):
            Directory containing TIFF image files.
        label_dir (Optional[str]):
            Directory containing label TIFF files.
        channel_map (Dict[str, str]):
            Mapping from filename codes to channel names.
        ignore (Optional[List[str]]):
            List of filename codes to ignore.
            Defaults to ["Merge"], which is a
            code for merged images.

    Returns:
        Dict[str, Dict[str, np.ndarray]]:
            A dictionary containing two keys: "images" and "labels".
            Each key maps to another dictionary where the keys are
            channel names and the values are numpy arrays of images.
    """
    # build a reference to the observations
    zstack_arrays = {
        "images": {
            channel_map.get(filename_code, f"Unknown_{filename_code}"): np.stack(
                [
                    tiff.imread(tiff_file.path).astype(np.uint16)
                    for tiff_file in sorted(
                        files,
                        key=lambda x: extract_z_slice_number_from_filename(x.name),
                    )
                ]
            ).astype(np.uint16)
            for filename_code, files in groupby(
                sorted(
                    [
                        file
                        for file in os.scandir(image_dir)
                        if (file.name.endswith(".tif") or file.name.endswith(".tiff"))
                        and (
                            file.name.split("_")[1] not in ignore
                            if ignore is not None
                            else True
                        )
                    ],
                    key=lambda x: x.name.split("_")[1],
                ),
                key=lambda x: x.name.split("_")[1],
            )
        }
    }

    if label_dir:
        zstack_arrays["labels"] = {
            f"{pathlib.Path(label_name).stem} (labels)": tiff.imread(
                next(iter(file)).path
            ).astype(np.uint16)
            for label_name, file in groupby(
                sorted(
                    [
                        file
                        for file in os.scandir(label_dir)
                        if (file.name.endswith(".tif") or file.name.endswith(".tiff"))
                        and (
                            file.name.split("_")[0] not in ignore
                            if ignore is not None
                            else True
                        )
                    ],
                    key=lambda x: x.name.split("_")[0],
                ),
                key=lambda x: x.name.split("_")[0],
            )
        }

    return zstack_arrays


def tiff_to_zarr(  # noqa: PLR0913
    image_dir: str,
    label_dir: Optional[str],
    output_path: str,
    channel_map: Dict[str, str],
    scaling_values: Union[List[int], Tuple[int]],
    ignore: Optional[List[str]] = ["Merge"],
) -> str:
    """
    Convert TIFF files to OME-Zarr format.

    Args:
        image_dir (str):
            Directory containing TIFF image files.
        label_dir (Optional[str]):
            Directory containing label TIFF files.
        output_path (str):
            Path to save the output OME-Zarr file.
        channel_map (Dict[str, str]):
            Mapping from filename codes to channel names.
        scaling_values (Union[List[int], Tuple[int]]):
            Scaling values for the images.
        ignore (Optional[List[str]]):
            List of filename codes to ignore.
            Defaults to ["Merge"], which is a
            code for merged images.

    Returns:
        str: Path to the output OME-Zarr file.
    """

    # except on dir already existing
    if pathlib.Path(output_path).is_dir():
        raise FileExistsError(
            (
                f"Output path {output_path} already exists."
                "Please remove before creating a new Zarr."
            )
        )

    if not pathlib.Path(image_dir).is_dir():
        raise NotADirectoryError(f"Image directory {image_dir} does not exist.")

    # build a reference to the observations
    frame_zstacks = image_set_to_arrays(
        image_dir=image_dir, label_dir=label_dir, channel_map=channel_map, ignore=ignore
    )

    # Parse URL and ensure store is compatible
    store = zarr_parse_url(output_path, mode="w").store
    # Ensure we are working with a Zarr group
    root = zarr.group(store, overwrite=True)

    # create scaling metadata
    scale_metadata = [
        {
            "datasets": [
                {
                    "path": "0",  # Path to the dataset
                    "coordinateTransformations": [
                        {
                            "type": "scale",
                            "scale": list(scaling_values),
                        }  # Apply scaling values
                    ],
                }
            ],
            "axes": [
                {
                    "name": "z",
                    "unit": "micrometer",
                    "type": "space",
                },  # Define the z-axis
                {
                    "name": "y",
                    "unit": "micrometer",
                    "type": "space",
                },  # Define the y-axis
                {
                    "name": "x",
                    "unit": "micrometer",
                    "type": "space",
                },  # Define the x-axis
            ],
        }
    ]

    # Write each channel separately to the Zarr file with no compression
    # Save images to OME-Zarr format
    images_group = root.create_group("images")
    for channel, stack in frame_zstacks["images"].items():
        zarr_write_image(
            image=stack,
            group=(group := images_group.create_group(channel)),
            axes="zyx",  # Specify the axes order for each channel
            dtype="uint16",  # Ensure the dtype is set correctly
            scaler=None,  # Disable scaler
        )
        # Set the units attribute for the group to "micrometers"
        group.attrs["units"] = "micrometers"

        # Define the multiscales metadata for the group
        group.attrs["multiscales"] = scale_metadata

    if label_dir:
        # Save masks to OME-Zarr format
        labels_group = root.create_group("labels")
        for compartment_name, stack in frame_zstacks["labels"].items():
            zarr_write_image(
                image=stack,
                group=(group := labels_group.create_group(compartment_name)),
                axes="zyx",  # Specify the axes order for each mask
                dtype="uint16",  # Ensure the dtype is set correctly
                scaler=None,  # Disable scaler
            )
            # Set the units attribute for the group to "micrometers"
            group.attrs["units"] = "micrometers"

            # Define the multiscales metadata for the group
            group.attrs["multiscales"] = scale_metadata

    return output_path


def tiff_to_ometiff(  # noqa: PLR0913
    image_dir: str,
    label_dir: Optional[str],
    output_path: str,
    channel_map: Dict[str, str],
    scaling_values: Union[List[int], Tuple[int]],
    ignore: Optional[List[str]] = ["Merge"],
) -> str:
    """
    Convert TIFF files to OME-TIFF format.

    Args:
        image_dir (str):
            Directory containing TIFF image files.
        label_dir (Optional[str]):
            Directory containing label TIFF files.
        output_path (str):
            Path to save the output OME-TIFF file.
        channel_map (Dict[str, str]):
            Mapping from filename codes to channel names.
        scaling_values (Union[List[int], Tuple[int]]):
            Scaling values for the images.
        ignore (Optional[List[str]]):
            List of filename codes to ignore.
            Defaults to ["Merge"], which is a
            code for merged images.

    Returns:
        str: Path to the output OME-TIFF file.
    """

    # except on dir already existing
    if pathlib.Path(output_path).is_file():
        raise FileExistsError(
            (
                f"Output path {output_path} already exists."
                "Please remove before creating a new OME-TIFF."
            )
        )

    if not pathlib.Path(image_dir).is_dir():
        raise NotADirectoryError(f"Image directory {image_dir} does not exist.")

    frame_zstacks = image_set_to_arrays(
        image_dir=image_dir, label_dir=label_dir, channel_map=channel_map, ignore=ignore
    )

    # Prepare the data for writing
    images_data = []
    labels_data = []
    channel_names = []
    label_names = []

    # Collect image data
    for channel, stack in frame_zstacks["images"].items():
        images_data.append(stack)
        channel_names.append(channel)

    # Collect label data
    if label_dir:
        for compartment_name, stack in frame_zstacks["labels"].items():
            labels_data.append(stack)
            label_names.append(compartment_name)

    # Stack the images and labels along a new axis for channels
    images_data = np.stack(images_data, axis=0)
    if labels_data:
        labels_data = np.stack(labels_data, axis=0)
        combined_data = np.concatenate((images_data, labels_data), axis=0)
        combined_channel_names = channel_names + label_names
    else:
        combined_data = images_data
        combined_channel_names = channel_names

    # Generate OME-XML metadata
    ome_metadata = {
        "SizeC": combined_data.shape[0],
        "SizeZ": combined_data.shape[1],
        "SizeY": combined_data.shape[2],
        "SizeX": combined_data.shape[3],
        "PhysicalSizeX": scaling_values[2],
        "PhysicalSizeY": scaling_values[1],
        "PhysicalSizeZ": scaling_values[0],
        # note: we use 7-bit ascii compatible characters below
        # due to tifffile limitations
        "PhysicalSizeXUnit": "um",
        "PhysicalSizeYUnit": "um",
        "PhysicalSizeZUnit": "um",
        "Channel": [{"Name": name} for name in combined_channel_names],
    }
    ome_xml = generate_ome_xml(ome_metadata)

    # Write the combined data to a single OME-TIFF
    with tiff.TiffWriter(output_path, bigtiff=True) as tif:
        tif.write(combined_data, description=ome_xml, photometric="minisblack")

    return output_path
